# -*- coding: utf-8 -*-
#Find missing issues using an offline comicvine comic list
#FindMissingIssues_CV.py
#
#Author: Xelloss Nakama
#
#Description:   This script loads an offline comic list built from the comicvine database and use it
#               to find missing comics in a selected group of comics in your library
#               (comics from the volumes of the selected comics, not present in the selected comics)
#               and then it create a list of these missing issues in a text file
#
#Versions:
#               0.1 First version (database date: 15/1/2017)
#               0.2 Fixed order of numbers in each volume, improved speed a bit (database date: 15/1/2017)
#               0.3 Change libraries dependencies to Windows Default instead of IronPython (better system compatibility), add option to search for input data file instead of using default
#               0.4 Change input data file extension to "mcl", added database date info in input file, added input file format verificarion (database date: 15/1/2017)
#               0.5 Separate data file input from script version (now each one have different versions)
#               0.6 Fixed problem with unicode characters (unicode always my nightmare XD) - Still some problems with unicode characters in results but no crashes anymore because of that
#               0.7 Fixed visual problem with large string numbers and series with no year
#               0.8 Now the results are sorted by year (the year is the one from the first comic in your library of that series)
#               0.9 Great improve in speed of analysis (what took 10 minutes before it takes 5 seconds now), also much "cleaner" code (less exceptions)
#               1.0 Added comicvine volume_id in the results
#               1.1 Fixed bug with comics with non volumes crashes
#               1.2 Result is now saved in Documents folder (fixing some permission problems in same systems)
#               2.0 Added feature to load comic checklists!
#               2.1 Added Summary in the end of the Report
#               2.2 Fixed a problem with Unscrapped comics in checklist that cause a crash in some cases
#               2.3 Added autocomplete custom values for comics that don't have them (but have the issue info in notes)
#               2.4 Fixed how unscrapped comics affected results
#               2.5 Fixed problems with '"', ',', ';' in comic numbers
#               2.6 Added changes for new convention in storing ";" and "," in the mcl files
#               2.7 Better results sorting (now by last comic in library date for series, and fixed decimal sorting for number). Also added last comic in library date to results. Fixed "Repeated Values" for mcl mode.
#               2.8 Added Icon for the script in Automation, and also button in Library Files and Library Buttons
#
#Feel free to use or modify any or all of this as you like.
#
#ComicRack Declarations
#
#@Name  Find Missing Issues Offline
#@Hook  Books, Library
#@Image FindMissingIssues_CV.png
#@Key   FindMissingIssues_CV
#@PCount 0
#
import clr
import System, sys

clr.AddReference('System.Windows.Forms')

from System.Windows.Forms import OpenFileDialog, MessageBox
from System.Diagnostics.Process import Start

reload(sys)
sys.setdefaultencoding('utf-8')

def FindMissingIssues_CV(books):

    # I ask for the Data File
    if books:
    
        FileOk = False
        while FileOk == False:
            file_dialog = System.Windows.Forms.OpenFileDialog()
            file_dialog.Filter = "Missing Comics Lists (*.mcl)|*.mcl|Missing Comics Checklists (*.mcc)|*.mcc"
            dialog_result = file_dialog.ShowDialog()

            if dialog_result == System.Windows.Forms.DialogResult.Cancel:
                MessageBox.Show("Please select a Missing Comics List file")
                break

            # I load the data file in memory

            comicvine_data = open(file_dialog.FileName, 'r')
            comicvine_lines = comicvine_data.readlines()

            if comicvine_lines[0].split(";")[0] == "Missing":
                FileOk = True
                Checklist = False
            elif comicvine_lines[0].split(";")[0] == "Checklist":
                FileOk = True
                Checklist = True
            else:
                MessageBox.Show("Corrupted of wrong file format...")

            comicvine_data.close()

        if FileOk:

            # I build the comics to analyse tree

            selected_books_tree = dict()
            series_to_check = dict()

            book_counter = 0
            noncv_counter = 0
            
            # I autocomplete comicvine custom values if not present
            
            if not Checklist:            
            
                volume_issue_flag = False
                
                for book in books:
                    if (book.GetCustomValue("comicvine_volume") == None) or (book.GetCustomValue("comicvine_issue") == None):
               
                        if "ComicVine [CVDB" in book.Notes:
                            
                            # I build the data dictionary for autopopulate custom values if needed.
                            
                            if not volume_issue_flag:
                            
                                issue_volume = dict()
                                volume_issue_flag = True
            
                                for line in comicvine_lines[1:]:
                                    
                                    line_data = line.split(";")
                                    
                                    data_issues = line_data[1].split(",")
                                    data_volume = line_data[0]
                                    
                                    for data_issue in data_issues:
                                    
                                        issue_volume[data_issue] = data_volume   
                                            
                            
                            comic_issue = book.Notes[book.Notes.find("ComicVine [CVDB")+15:book.Notes.find("] on")]
                            
                            if issue_volume.has_key(comic_issue):
                            
                                comic_volume =  issue_volume[comic_issue]
                                book.SetCustomValue("comicvine_issue",comic_issue)
                                book.SetCustomValue("comicvine_volume",comic_volume)
           
            
            for book in books:

                # I load the data from the comics

                volume = unicode(book.GetCustomValue("comicvine_volume"))
                issue = unicode(book.GetCustomValue("comicvine_issue"))
                name = unicode(book.Series) + " (" + unicode(book.Volume) + ") " + "[" + unicode(volume) + "]"
                
                book_counter += 1

                # I build the tree

                if (book.GetCustomValue("comicvine_volume") == None) or (book.GetCustomValue("comicvine_issue") == None):
                    noncv_counter += 1
                    
                else:
                
                    if selected_books_tree.has_key(volume):

                        selected_books_tree[volume][issue] = "OK"

                    else:

                        selected_books_tree[volume] = dict()
                        selected_books_tree[volume]["name"] = name
                        selected_books_tree[volume][issue] = "OK"
                        
                    if not Checklist:
                        if series_to_check.has_key(volume):
                            if series_to_check[volume] < unicode(book.Year) + "/" + unicode(book.Month) + "-" + book.GetCustomValue("comicvine_issue"):
                                series_to_check[volume] = unicode(book.Year) + "/" + unicode(book.Month) + "-" + book.GetCustomValue("comicvine_issue")
                        else:
                            series_to_check[volume] = unicode(book.Year) + "/" + unicode(book.Month) + "-" + book.GetCustomValue("comicvine_issue")

            # I build the comicvine database comics tree with the data I need

            comicvine_books_tree = dict()

            for line in comicvine_lines[1:]:

                # I parse the data from each text line

                fix = False
                line.replace(';"', ';')
                if line[len(line)-1] == '"':
                    line = line[0:-2]
                    fix = True
                
                line_data = line.split(";")
                
                if fix:            
                    line_data[2] = line_data[2].replace(', ,',',-,')
                    line_data[2] = line_data[2].replace(', ',' ')
                                    
                data_issues = line_data[1].split(",")
                data_volume = line_data[0]            
                data_numbers = line_data[2].split(",")

                if Checklist:
                    series_to_check[data_volume] = line_data[3]
                    

                # I load the comics only of the comics of the volumes I selected

                # I see if I need the volume

                if selected_books_tree.has_key(data_volume) or Checklist:

                    # I build the tree

                    comicvine_books_tree[data_volume] = []
                    
                    for i in range(0,len(data_issues)):

                        data_pair = []

                        data_pair.append(data_issues[i])
                        data_pair.append(data_numbers[i])

                        comicvine_books_tree[data_volume].append(data_pair)

            # I begin comparing data between the selected comics and the comics loaded from the database and print the differences in the output file

            count = 0

            # I select the output file

            documents = System.Environment.GetFolderPath(System.Environment.SpecialFolder.MyDocuments);
            outputfile = open(documents+'/output.txt', 'w')

            if Checklist:
                outputfile.write(comicvine_lines[0].split(";")[2] + " CheckList - Updated: " + comicvine_lines[0].split(";")[1] + "\n")
                for letters in comicvine_lines[0].split(";")[2]:
                    outputfile.write("=")
                outputfile.write("================================\n\n")
            else:
                outputfile.write("Missing issues till " + comicvine_lines[0].split(";")[1])
                outputfile.write("================================\n\n")

            missing_comics_counter = 0
            incomplete_series_counter = 0
            match_comics_counter = 0

            for comic_series in sorted(series_to_check.keys(), key=lambda serie: series_to_check[serie]):
                
                miss_flag = False
                
                if comicvine_books_tree.has_key(comic_series):
                    
                    if Checklist:
                    
                        if not selected_books_tree.has_key(comic_series):
                            
                            selected_series = dict()
                        
                        else:
                        
                            selected_series = selected_books_tree[comic_series]
                    
                    else:
                        
                        selected_series = selected_books_tree[comic_series]

                    comicvine_books_in_series = comicvine_books_tree[comic_series]

                    missing_list = []

                    # I search for missing numbers in each volume

                    for comic in comicvine_books_in_series:

                        if selected_series.has_key(comic[0]):

                            # If the comic exist, I delete it from the selected list

                            del selected_books_tree[comic_series][comic[0]]
                            match_comics_counter += 1

                        else:

                            # if not I add it to the missing list

                            miss_flag = True
                            missing_list.append(comic)

                    # If there are missing numbers, I create an output entry

                    if miss_flag == True:

                        incomplete_series_counter += 1
                        
                        if count != 0:
                            outputfile.write("\n")

                        if Checklist:
                            outputfile.write("\n" + series_to_check[comic_series].encode('utf-8','ignore') + " [" +comic_series+ "]:\n")
                        else:
                            outputfile.write("\n" + selected_books_tree[comic_series]["name"].encode('utf-8','ignore') + " - (Last " + series_to_check[comic_series].split("-")[0] + "):\n")

                        # I print all the missing numbers

                        issues_list = ""

                        for missing_comic in missing_list:

                             # I give format to the comic numbers

                            missing_comic[1] = missing_comic[1].replace(".&@1",",")
                            missing_comic[1] = missing_comic[1].replace(".&@2",";")
                            missing_comic[1] = missing_comic[1].replace("\n","")
                            missing_comic[1] = missing_comic[1].replace("\r","")
                            missing_comic[1] = missing_comic[1].replace(" ","")

                            try:

                                missing_comic[1] = str(int(missing_comic[1])).zfill(3)

                            except:

                                try: 
                                 
                                    missing_comic[1] = str(int(missing_comic[1].split(".")[0])).zfill(3) + missing_comic[1][missing_comic[1].find("."):]

                                except:
                                
                                    a = 0

                                    
                            if len(missing_comic[1]) == 3:
                                issues_list += "  "
                            if len(missing_comic[1]) == 4:
                                issues_list += " "
                            if len(missing_comic[1]) == 1:
                                issues_list += "    "
                            if len(missing_comic[1]) == 2:
                                issues_list += "   "

                            # I print the numbers in a list

                            issues_list += missing_comic[1] + "&12&"


                        issues_list2 = sorted(issues_list.split("&12&"))

                        count = 0

                        for number in issues_list2[1:]:
                        
                            count += 1
                            missing_comics_counter += 1

                            if count > 15:
                                count = 1
                                outputfile.write("\n")

                            if len(number) > 5:

                                if len(number) == 6:
                                    count += 1
                                    number += "     "

                                if len(number) == 7:
                                    count += 1
                                    number += "    "

                                if len(number) == 8:
                                    count += 1
                                    number += "   "

                                if len(number) == 9:
                                    count += 1
                                    number += "  "

                                if len(number) == 10:
                                    count += 1
                                    number += " "

                            outputfile.write(number.encode('utf-8','ignore').replace("Â","") + " ")   
            
            if Checklist:
                
                outputfile.write("\n\n\nSelected series not in collection:\n==================================")
                
                redundant_comics_counter = 0
                
                for redundant_series in sorted(selected_books_tree.keys()):
                
                    if len(selected_books_tree[redundant_series])>1:
                    
                        outputfile.write("\n" + selected_books_tree[redundant_series]["name"].encode('utf-8','ignore') + " [" +redundant_series+ "]")
                        
                        redundant_comics_counter += (len(selected_books_tree[redundant_series])-1)
                    
                
                outputfile.write("\n\n\nSummary:\n=======+ \nSelected Comics: " + str(book_counter) + "\nMissing comics: " + str(missing_comics_counter)+ " (in "+ str(incomplete_series_counter))
                outputfile.write(" Series)\nMatched comics: " + str(match_comics_counter) + " (" + str(int(match_comics_counter*10000.0/(match_comics_counter + missing_comics_counter))/100.0)+"% of Total)")
                outputfile.write("\nComics not in collection: " + str(redundant_comics_counter) + "\nUnscrapped Comics: " + str(noncv_counter))
                outputfile.write("\nRepeated Comics: " + str(book_counter-match_comics_counter-redundant_comics_counter-noncv_counter))
                
            else:
                
                redundant_comics_counter = 0
                ids = ""
                
                for redundant_series in sorted(selected_books_tree.keys()):
                
                    if len(selected_books_tree[redundant_series])>1:
                    
                        for id in selected_books_tree[redundant_series]:
                            
                                ids += "-" + str(id)
                        
                        redundant_comics_counter += (len(selected_books_tree[redundant_series])-1)
                        
                outputfile.write("\n\n\nSummary:\n=======\nSelected Comics: " + str(book_counter) + "\nMissing comics: " + str(missing_comics_counter) + " (in "+ str(incomplete_series_counter) + " Series)\nMatched comics: " + str(match_comics_counter))
                outputfile.write("\nComics not in Database: " + str(redundant_comics_counter) + " : " + ids +"\nUnscrapped Comics: " + str(noncv_counter) + "\nRepeated Comics: " + str(book_counter-match_comics_counter-noncv_counter))
            
            outputfile.close()

            # I open the output file

            Start("notepad", documents + "/output.txt");